# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Chris Rizzitello <crizzitello@ics.com>
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

from ics_sbom_libs.common.vulnerability import Vulnerability


class MatchResult:

    def __init__(self, name: str, version: str, cpes: list[str] = None):
        self.cve_list: list[Vulnerability] = []
        self.name = name
        self.version = version
        self.cpe_list: list[str] = cpes if cpes else []

    def append_cve(self, cve: Vulnerability):
        if cve not in self.cve_list:
            self.cve_list.append(cve)

    def get_severity_info(self):
        severity_counts = {"NONE": 0, "LOW": 0, "MEDIUM": 0, "HIGH": 0, "CRITICAL": 0}
        for item in self.cve_list:
            severity_counts[item.severity] += 1

        return severity_counts

    @property
    def stringify(self):
        line = f'{self.name} ({self.version})[{" ".join(self.cpe_list)}]: Has {len(self.cve_list)} CVEs'
        severity_info = self.get_severity_info()
        for severity in severity_info.keys():
            line += f" {severity}: {severity_info[severity]}"
        if not self.cve_list:
            line += "!"
        else:
            line += ": " + " ".join((item.cve_number for item in self.cve_list))
        return line

    @property
    def csvify(self):
        out = f'{self.name},{self.version},{" ".join(self.cpe_list)},{len(self.cve_list)},'
        severity_info = self.get_severity_info()
        for severity in severity_info.keys():
            out += f"{severity_info[severity]},"
        if self.cve_list:
            out = out + f'{" ".join((item.cve_number for item in self.cve_list))}'
        return out

    def print_result(self):
        print(self.stringify)

    def __lt__(self, other):
        return self.name < other.name



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Sergey Missan <smissan@ics.com>
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>
# SPDX-FileContributor: Chris Rizzitello <crizzitello@ics.com>
# SPDX-FileContributor: Qin Zhang <qzhang@ics.com>
# SPDX-FileContributor: Boris Ralchenko <bralchenko@ics.com>

import sqlite3
import time
import json
import os
import requests
import argparse

from typing import Final
from pathlib import Path
from datetime import datetime
from cpeparser import CpeParser
from tqdm import tqdm

from ics_sbom_libs.common.vulnerability import Vulnerability

# Setup Logging
import logging

log = logging.getLogger(__name__)

FETCH_INTERVAL_SEC: Final = 2
FETCH_INTERVAL_SEC_NO_API_KEY: Final = 6
CVE_RECORD_PER_PAGE: Final = 2000
CPE_RECORD_PER_PAGE: Final = 10000
# CVE v2. See https://nvd.nist.gov/developers/vulnerabilities for refrence
NIST_BASE_URL: Final = "https://services.nvd.nist.gov/rest/json/{}/2.0/"
CVE_URL: Final = NIST_BASE_URL.format("cves")
CPE_URL: Final = NIST_BASE_URL.format("cpes")

_cache_dir = Path("~").expanduser() / ".cache" / "icsbom"

# see cve_schema.sql for the version numbers.
_nvd_db_version = "4.0"
_nvd_dbFile = f"nvd_v{_nvd_db_version}.db"
_api_key = ""  # this will for it to load the saved api_key if one is saved.


class VulnerabilityDatabase:
    _cache_dir: Path
    _db_file_name: str
    _api_key: str | None

    def __init__(self, cache_dir=_cache_dir, db_file=_nvd_dbFile, api_key=_api_key):
        self.cache_dir = cache_dir
        self.db_file_name = db_file
        self.api_key = api_key

        self._last_query: datetime | None = None

        self._setup()

    def __del__(self):
        if self.con:
            self.con.close()

    @property
    def cache_dir(self):
        return self._cache_dir

    @cache_dir.setter
    def cache_dir(self, new_dir):
        if new_dir is None:
            return

        if isinstance(new_dir, str):
            self._cache_dir = Path(new_dir)
        elif isinstance(new_dir, Path):
            self._cache_dir = new_dir

        # Handle the possibility that '~' was used in the path.
        self._cache_dir = self._cache_dir.expanduser()

        # In the event there are any symbolic links in the path, let's just resolve everything out.
        self._cache_dir = self._cache_dir.resolve()

    @property
    def db_file_name(self):
        return self._db_file_name

    @db_file_name.setter
    def db_file_name(self, new_file):
        if isinstance(new_file, str):
            self._db_file_name = new_file

    @property
    def api_key(self):
        return self._api_key

    @api_key.setter
    def api_key(self, new_key):
        if not new_key:
            self._api_key = self._load_api_key()
            return

        if isinstance(new_key, str):
            if new_key == "none":
                self._api_key = ""
            else:
                self._api_key = new_key

    def _load_api_key(self):
        if not self.cache_dir or not self.cache_dir.exists():
            return ""

        api_key_path = self.cache_dir / "api_key.txt"
        if not api_key_path.exists():
            return ""

        api_key = api_key_path.read_text().strip()
        if not api_key:
            return ""

        return api_key

    def _save_api_key(self):
        if not self.cache_dir or not self.cache_dir.exists() or not self.api_key:
            return

        api_key_path = self.cache_dir / "api_key.txt"
        api_key_path.write_text(f"{self.api_key}")

    @property
    def db_path(self):
        return self.cache_dir / self.db_file_name

    def _setup(self):
        if not self.cache_dir.exists():
            self.cache_dir.mkdir(parents=True)

        if not os.path.exists(self.db_path):
            self._init_db_()

        self.con = sqlite3.connect(self.db_path)

    @staticmethod
    def setup_args(parser: argparse.ArgumentParser):
        if not parser:
            return

        parser.add_argument("--cache_dir", type=str, default=_cache_dir, help="Cache Directory: contains the database")
        parser.add_argument("--api_key", type=str, default=_api_key, required=False, help="API key for NVD database")
        parser.add_argument(
            "--save_key",
            action="store_true",
            required=False,
            help="Save the API key for the NVD database to a cache file in cache_dir.",
        )
        parser.add_argument(
            "--db_file",
            type=str,
            default=f"{_nvd_dbFile}",
            help="Database file path. If file does not exist, a new database will be generated "
            "from scratch and saved to the given path. This will take sometime.",
        )

    def process_args(self, args):
        if not args:
            return

        self.cache_dir = args.cache_dir
        self.db_file_name = args.db_file
        self.api_key = args.api_key

        if args.save_key and args.api_key != "none":
            self._save_api_key()
            args.save_key = False

        self._setup()

    def _init_db_(self):
        with open(Path(__file__).parent.resolve() / "cve_schema.sql") as fp:
            self.con = sqlite3.connect(self.db_path)
            cur = self.con.cursor()
            cur.executescript(fp.read())

    def _get_latest_timestamp_(self):
        try:
            curs = self.con.execute("SELECT last_modified FROM cve_severity ORDER BY last_modified DESC LIMIT 1;")
            latest = curs.fetchone()[0]
            curs.close()
            return latest
        except Exception:
            return None

    def _get_status_value(self, key: str, default=None):
        try:
            curs = self.con.execute("SELECT value FROM status WHERE key = ?", (key,))
            item: str = curs.fetchone()
            curs.close()
            if item is None:
                return default
            return item[0]
        except Exception:
            return default

    def _set_status_value(self, key: str, value):
        try:
            curs = self.con.cursor()
            curs.execute("REPLACE INTO status (key, value) VALUES (?,?)", (key, f"{value}"))
            self.con.commit()
        except Exception:
            return

    def _remove_status_key(self, key: str):
        try:
            curs = self.con.cursor()
            curs.execute("DELETE FROM status WHERE key = ?", (key,))
            self.con.commit()
        except Exception:
            return

    def _query_nvd(self, nvd_url: str, query: str):
        query_url = f"{nvd_url}?"

        query_interval = FETCH_INTERVAL_SEC if self.api_key else FETCH_INTERVAL_SEC_NO_API_KEY
        query_delta = (datetime.now() - self._last_query).seconds if self._last_query else query_interval + 1
        if query_delta < query_interval:
            time.sleep(query_interval - query_delta)

        if self.api_key:
            response = requests.get(query_url + query, headers={"apiKey": self.api_key})
        else:
            response = requests.get(query_url + query)

        self._last_query = datetime.now()

        return response

    def query_cve_from_nvd(self, cve_id: str):
        if not cve_id:
            return None

        cve_query = f"cveId={cve_id}"
        retry_count = 0
        finished = False
        data = None
        while True:

            try:
                response = self._query_nvd(CVE_URL, cve_query)
            except Exception as e:
                log.exception(f"{e}")
                break

            if response.status_code == 503:
                time.sleep(6)
                retry_count += 1
                continue
            else:
                retry_count = 0

            try:
                data = response.text
                finished = True

            except Exception as e:
                log.exception(str(e) + "\n" + "\nstatus_code:" + str(response.status_code))
                data = None
                break

            if finished:
                break

        return data

    def create_database(self):

        def get_data(src_name: str, src_url, data_handler, records_per_page):
            # Grab a minimal amount of data just to get the totals.
            download_db = not bool(self._get_status_value(f"initial_{src_name}_download_completed", False))

            start_index = int(self._get_status_value(f"last_{src_name}_record_received", 0)) if download_db else 0
            last_update = self._get_status_value(f"{src_name}_last_updated")
            last_update_query = ""
            if not download_db and last_update is not None:
                last_update = last_update.replace("'", "")
                diff = datetime.utcnow() - datetime.fromisoformat(last_update)
                if diff.days > 100:
                    self._init_db_()  # Not allow to have long duration data fetch
                else:
                    log.info(f"Updating {src_name.upper()} records since last update: {last_update}")
                    last_update_query = f"lastModStartDate={last_update}&lastModEndDate=" + str(
                        datetime.utcnow().isoformat()
                    )

            pbar = tqdm(total=1, desc=f"Updating NVD {src_name.upper()} Database", unit_scale=1, unit="Records")
            pbar.update(start_index)
            retry_count = 0
            finished = False
            while True:

                try:
                    response = self._query_nvd(
                        src_url, last_update_query + f"&resultsPerPage={records_per_page}&startIndex={start_index}"
                    )
                except Exception as e:
                    log.exception(f"{e}")
                    break

                if response.status_code == 503:
                    pbar.set_postfix_str("Request Failed...")
                    time.sleep(6)
                    retry_count += 1
                    pbar.set_postfix_str(f"Request Failed... Retrying (Retry Count: {retry_count})")
                    continue
                else:
                    pbar.set_postfix_str(refresh=False)
                    retry_count = 0

                try:
                    data = json.loads(response.text)
                    data_handler(data)
                    if data["startIndex"] + data["resultsPerPage"] >= data["totalResults"]:
                        finished = True

                    else:
                        start_index = data["startIndex"] + data["resultsPerPage"]

                except Exception as e:
                    log.exception(str(e) + "\n" + "\nstatus_code:" + str(response.status_code))
                    break

                if pbar.total != data["totalResults"]:
                    pbar.total = data["totalResults"]
                    pbar.refresh()

                # self.con.commit()

                if not finished:
                    self._set_status_value(f"last_{src_name}_record_received", start_index)
                else:
                    self._remove_status_key(f"last_{src_name}_record_received")
                    self._set_status_value(f"initial_{src_name}_download_completed", finished)

                self._set_status_value(f"{src_name}_last_updated", datetime.utcnow().isoformat())

                self.con.commit()
                pbar.update(data["resultsPerPage"])

                if finished:
                    break

            pbar.close()

        get_data("cve", CVE_URL, self._process_cve_data_, CVE_RECORD_PER_PAGE)
        get_data("cpe", CPE_URL, self._process_cpe_data_, CPE_RECORD_PER_PAGE)

        self.con.close()

    def _process_cve_data_(self, data):
        cveArray = []
        rangeArray = []
        weaknessArray = []
        cves = data["vulnerabilities"]
        for cve in cves:
            ranges = []
            weakness = []
            config_block = []
            data = CveDataHelper.get_data(cve["cve"], ranges, weakness, config_block)

            if data is not None:
                cveArray.append(data)
                if len(ranges) > 0:
                    rangeArray.extend(ranges)
                if len(weakness) > 0:
                    weaknessArray.extend(weakness)

        curs = self.con.cursor()
        vuln_names = Vulnerability.sql_query_name_list()
        curs.executemany(
            "INSERT INTO cve_severity (" + ", ".join(vuln_names) + ") "
            "VALUES (" + ", ".join(f":{name}" for name in vuln_names[0:-1]) + ", 'NVD') "
            "ON CONFLICT (cve_number) DO UPDATE SET " + ", ".join(f"{name}=:{name}" for name in vuln_names[1:-1]) + ";",
            cveArray,
        )

        # Clear ranges and weakness for CVE records
        cveNums = list()
        for item in cveArray:
            cveNums.append("'" + item["cve_number"] + "'")
        cveNums = ",".join(cveNums)
        curs.execute(f"DELETE FROM cve_range WHERE cve_number IN ({cveNums})")
        curs.execute(f"DELETE FROM cve_weakness WHERE cve_number IN ({cveNums})")

        # TODO, limit duplicated ranges.
        if len(rangeArray) > 0:
            key_list = [
                "cve_number",
                "vendor",
                "product",
                "version",
                "cpe",
                "vulnerable",
                "versionStartIncluding",
                "versionStartExcluding",
                "versionEndIncluding",
                "versionEndExcluding",
                "part_type",
                "data_source",
            ]
            sql_str = (
                "INSERT INTO cve_range ("
                + ", ".join(key_list)
                + ") "
                + " VALUES ("
                + ", ".join(f":{name}" for name in key_list[0:-2])
                + ", :part , 'NVD') "
            )
            curs.executemany(sql_str, rangeArray)

        if len(weaknessArray) > 0:
            curs.executemany(
                "INSERT INTO cve_weakness (cve_number, value) VALUES (:cve_number, :value) "
                "ON CONFLICT (cve_number, value) DO NOTHING",
                weaknessArray,
            )

    def _process_cpe_data_(self, data):
        cpeArray = []
        products = data["products"]
        for product in products:
            # Are there multiple products in the CWE NVD table?
            if "cpe" in product.keys():
                cpeArray.append(CpeRecordDataHelper.parse_data(product["cpe"]))

        if not cpeArray:
            return

        curs = self.con.cursor()
        cpe_properties = list(cpeArray[0].keys())
        curs.executemany(
            "INSERT INTO cpe_dictionary ("
            + ", ".join(cpe_properties)
            + ") "
            + "VALUES ("
            + ", ".join(f":{name}" for name in cpe_properties)
            + ") "
            + "ON CONFLICT (cpe_id) DO UPDATE SET "
            + ", ".join(f"{name}=:{name}" for name in cpe_properties[1:])
            + ";",
            cpeArray,
        )

    def query_cache(self, query: str, parameters=()):
        if not query:
            return None

        return self.con.execute(query, parameters)

    def query_cpe_dictionary(self, package_name: str):
        query = f"SELECT cpe FROM cpe_dictionary WHERE product='{package_name}'"
        cpe_strings = []

        cursor = self.con.execute(query)
        if not cursor:
            return cpe_strings

        res = cursor.fetchall()
        if not res:
            return cpe_strings

        for cpe in res:
            cpe_strings.append(cpe[0])

        return cpe_strings

    def get_cve(self, cve_id: str):
        if not self.con:
            return None

        results = self.query_cache(
            "SELECT " + ", ".join(Vulnerability.sql_query_name_list()) + " FROM cve_severity WHERE cve_number=?",
            (cve_id,),
        )
        cve = Vulnerability(results.fetchone())

        results = self.query_cache("SELECT value FROM cve_weakness WHERE cve_number=?", (cve_id,))
        cve.cwes = [cwe[0] for cwe in results.fetchall()]

        return cve


class CveDataHelper:
    range_fields: Final = [
        "versionStartIncluding",
        "versionStartExcluding",
        "versionEndIncluding",
        "versionEndExcluding",
    ]

    @staticmethod
    def get_data(cve, ranges, weakness, config_block):
        desc = cve["descriptions"][0]["value"]
        if desc.startswith("** REJECT **") or desc.startswith("Rejected reason:"):
            return None
        cve_number = cve["id"]
        dateTime = cve["lastModified"] if "lastModified" in cve else cve["published"]
        data = {"cve_number": cve_number, "description": desc, "last_modified": dateTime}
        CveDataHelper.get_metrics(cve, data)
        CveDataHelper.get_range(cve, cve_number, ranges)
        CveDataHelper.get_configurations(cve, config_block)
        CveDataHelper.get_weakness(cve, cve_number, weakness)
        data["configurations"] = json.dumps(config_block)

        return data

    @staticmethod
    def parse_node(cve, cve_number) -> []:
        ret = []
        if "cpeMatch" not in cve:
            return ret
        cpe = CpeParser()

        for cpeMatch in cve["cpeMatch"]:
            result = cpe.parser(cpeMatch["criteria"])
            result["cve_number"] = cve_number
            result["cpe"] = cpeMatch["criteria"]
            result["vulnerable"] = cpeMatch["vulnerable"]
            for field in CveDataHelper.range_fields:
                if field in cpeMatch:
                    result[field] = cpeMatch[field]
                else:
                    result[field] = ""
            ret.append(result)
        return ret

    @staticmethod
    def get_weakness(cve, cve_number, weakness) -> None:
        if "weaknesses" not in cve:
            return

        for item in cve["weaknesses"]:
            weakness.append({"value": item["description"][0]["value"], "cve_number": cve_number})

    @staticmethod
    def get_range(cve, cve_number, ranges):
        if "configurations" not in cve:
            return

        for item in cve["configurations"]:  # TODO Check the configurations array. And the operator in the array.
            for node in item["nodes"]:
                ranges.extend(CveDataHelper.parse_node(node, cve_number))

    @staticmethod
    def get_configurations(cve, config_block):
        if "configurations" not in cve:
            return

        config_block.extend(cve["configurations"])

    @staticmethod
    def get_metrics(cve, data):
        if "cvssMetricV31" in cve["metrics"]:
            data["severity"] = cve["metrics"]["cvssMetricV31"][0]["cvssData"]["baseSeverity"]
            data["score"] = cve["metrics"]["cvssMetricV31"][0]["cvssData"]["baseScore"]
            data["cvss_vector"] = cve["metrics"]["cvssMetricV31"][0]["cvssData"]["vectorString"]
            data["cvss_version"] = float(cve["metrics"]["cvssMetricV31"][0]["cvssData"]["version"])
        elif "cvssMetricV30" in cve["metrics"]:
            data["severity"] = cve["metrics"]["cvssMetricV30"][0]["cvssData"]["baseSeverity"]
            data["score"] = cve["metrics"]["cvssMetricV30"][0]["cvssData"]["baseScore"]
            data["cvss_vector"] = cve["metrics"]["cvssMetricV30"][0]["cvssData"]["vectorString"]
            data["cvss_version"] = float(cve["metrics"]["cvssMetricV30"][0]["cvssData"]["version"])
        elif "cvssMetricV2" in cve["metrics"]:
            data["severity"] = cve["metrics"]["cvssMetricV2"][0]["baseSeverity"]
            data["score"] = cve["metrics"]["cvssMetricV2"][0]["cvssData"]["baseScore"]
            data["cvss_vector"] = cve["metrics"]["cvssMetricV2"][0]["cvssData"]["vectorString"]
            data["cvss_version"] = float(cve["metrics"]["cvssMetricV2"][0]["cvssData"]["version"])
        else:
            data["severity"] = "unknown"
            data["score"] = None
            data["cvss_vector"] = "unknown"
            data["cvss_version"] = 0


class CpeRecordDataHelper:
    property_map: Final = {
        "cpe_id": "cpeNameId",
        "cpe": "cpeName",
        "deprecated": "deprecated",
        "created": "created",
        "last_modified": "lastModified",
    }

    @staticmethod
    def parse_data(data):
        cpe_record = CpeRecordDataHelper.parse_cpe(data[CpeRecordDataHelper.property_map["cpe"]])
        for prop in CpeRecordDataHelper.property_map:
            cpe_record[prop] = data[CpeRecordDataHelper.property_map[prop]]

        return cpe_record

    @staticmethod
    def parse_cpe(cpe_name):
        result = CpeParser().parser(cpe_name)
        return {"vendor": result["vendor"], "product": result["product"]}



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

from cpeparser import CpeParser
from ics_sbom_libs.common.vulnerability import Vulnerability


class CpeMatchResult:

    def __init__(self, cpe: str):
        self._cpe = cpe
        self._parsed = CpeParser().parser(self._cpe)
        self._cve_list: list[Vulnerability] = []

    def append_cve(self, cve: Vulnerability):
        if cve not in self._cve_list:
            self._cve_list.append(cve)

    @property
    def product(self):
        return self._parsed["product"]

    @property
    def vendor(self):
        return self._parsed["vendor"]

    @property
    def version(self):
        return self._parsed["version"]

    @property
    def cve_list(self):
        return self._cve_list

    @property
    def cpe(self):
        return self._cpe

    def get_cpe_properties(self):
        return self.vendor, self.product, self.version



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

import unittest
import logging
import io

import argparse

from ics_sbom_libs.common import logging_setup


class TestLogging(unittest.TestCase):
    def test_logging(self):
        with self.assertLogs("foo", level="INFO") as cm:
            logging.getLogger("foo").info("first message")
            logging.getLogger("foo.bar").error("second message")
        self.assertEqual(cm.output, ["INFO:foo:first message", "ERROR:foo.bar:second message"])


def _log_level_tester(level: int | str):

    level_name: str = ""
    if isinstance(level, str):
        level_name = level
    else:
        level_name = logging.getLevelName(level)

    log_stream = io.StringIO()
    logging.basicConfig(level=level, stream=log_stream, force=True)

    log = logging.getLogger(level_name.lower() + "_test")

    log.debug("debug message")
    log.info("info message")
    log.warning("warning message")
    log.error("error message")
    log.critical("critical message")

    value = log_stream.getvalue()

    return value


class TestLogLevels(unittest.TestCase):

    def test_debug_log(self):
        test_level = logging.DEBUG
        test_name = logging.getLevelName(test_level).lower()
        result = _log_level_tester(test_level)

        self.assertEqual(
            result,
            f"DEBUG:{test_name}_test:debug message\n"
            f"INFO:{test_name}_test:info message\n"
            f"WARNING:{test_name}_test:warning message\n"
            f"ERROR:{test_name}_test:error message\n"
            f"CRITICAL:{test_name}_test:critical message\n",
        )

    def test_info_log(self):
        test_level = logging.INFO
        test_name = logging.getLevelName(test_level).lower()
        result = _log_level_tester(test_level)

        self.assertEqual(
            result,
            f"INFO:{test_name}_test:info message\n"
            f"WARNING:{test_name}_test:warning message\n"
            f"ERROR:{test_name}_test:error message\n"
            f"CRITICAL:{test_name}_test:critical message\n",
        )

    def test_warning_log(self):
        test_level = logging.WARNING
        test_name = logging.getLevelName(test_level).lower()
        result = _log_level_tester(test_level)

        self.assertEqual(
            result,
            f"WARNING:{test_name}_test:warning message\n"
            f"ERROR:{test_name}_test:error message\n"
            f"CRITICAL:{test_name}_test:critical message\n",
        )

    def test_error_log(self):
        test_level = logging.ERROR
        test_name = logging.getLevelName(test_level).lower()
        result = _log_level_tester(test_level)

        self.assertEqual(
            result, f"ERROR:{test_name}_test:error message\n" f"CRITICAL:{test_name}_test:critical message\n"
        )


class TestArgParsedLogLevels(unittest.TestCase):

    @staticmethod
    def _parse_args(args):
        parser = argparse.ArgumentParser()
        logging_setup.setup_log_arg(parser)

        return parser.parse_args(args)

    def test_debug_log(self):
        test_level = logging.DEBUG
        test_name = logging.getLevelName(test_level).lower()

        args = self._parse_args(["--log", test_name])
        result = _log_level_tester(logging_setup.handle_log_arg(args))

        self.assertEqual(
            result,
            f"DEBUG:{test_name}_test:debug message\n"
            f"INFO:{test_name}_test:info message\n"
            f"WARNING:{test_name}_test:warning message\n"
            f"ERROR:{test_name}_test:error message\n"
            f"CRITICAL:{test_name}_test:critical message\n",
        )

    def test_debug_log_2(self):
        test_level = logging.DEBUG
        test_name = logging.getLevelName(test_level).lower()

        args = self._parse_args([f"--log={test_name}"])
        result = _log_level_tester(logging_setup.handle_log_arg(args))

        self.assertEqual(
            result,
            f"DEBUG:{test_name}_test:debug message\n"
            f"INFO:{test_name}_test:info message\n"
            f"WARNING:{test_name}_test:warning message\n"
            f"ERROR:{test_name}_test:error message\n"
            f"CRITICAL:{test_name}_test:critical message\n",
        )

    def test_info_log(self):
        test_level = logging.INFO
        test_name = logging.getLevelName(test_level).lower()

        args = self._parse_args(["--log", test_name])
        result = _log_level_tester(logging_setup.handle_log_arg(args))

        self.assertEqual(
            result,
            f"INFO:{test_name}_test:info message\n"
            f"WARNING:{test_name}_test:warning message\n"
            f"ERROR:{test_name}_test:error message\n"
            f"CRITICAL:{test_name}_test:critical message\n",
        )

    def test_warning_log(self):
        test_level = logging.WARNING
        test_name = logging.getLevelName(test_level).lower()

        args = self._parse_args(["--log", test_name])
        result = _log_level_tester(logging_setup.handle_log_arg(args))

        self.assertEqual(
            result,
            f"WARNING:{test_name}_test:warning message\n"
            f"ERROR:{test_name}_test:error message\n"
            f"CRITICAL:{test_name}_test:critical message\n",
        )

    def test_error_log(self):
        test_level = logging.ERROR
        test_name = logging.getLevelName(test_level).lower()

        args = self._parse_args(["--log", test_name])
        result = _log_level_tester(logging_setup.handle_log_arg(args))

        self.assertEqual(
            result, f"ERROR:{test_name}_test:error message\n" f"CRITICAL:{test_name}_test:critical message\n"
        )


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright 2024 Ics inc.

from .base import VersionHandler, invalid_version_list
from .buildversion import BuildVersion
from .versionfactory import VersionFactory
from . import handlers

__all__ = ["VersionHandler", "invalid_version_list", "BuildVersion", "VersionFactory", "handlers"]



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>
# SPDX-FileContributor: Gerardo Stola <gstola@ics.com>

from datetime import datetime, tzinfo, timedelta
import textwrap

import rich.padding
from rich.text import Text
from rich.color import Color
from rich.style import Style


class SimpleUtc(tzinfo):
    """
    derived from
    https://stackoverflow.com/questions/19654578/
       python-utc-datetime-objects-iso-format-doesnt-include-z-zulu-or-zero-offset
    """

    def tzname(self, **kwargs):
        return "UTC"

    def utcoffset(self, dt):
        return timedelta(0)

    def dst(self, __dt):
        return 0


class SeverityProperties:
    priority: int
    style: str | Style
    indicator: str

    def __init__(self, priority: int, style: str | Style, indicator: str):
        self.priority = priority
        self.style = style
        self.indicator = indicator

    def __lt__(self, other):
        if isinstance(other, SeverityProperties):
            return self.priority < other.priority

        return False

    def __gt__(self, other):
        if isinstance(other, SeverityProperties):
            return self.priority > other.priority

        return False

    def __le__(self, other):
        if isinstance(other, SeverityProperties):
            return self.priority <= other.priority

        return False

    def __ge__(self, other):
        if isinstance(other, SeverityProperties):
            return self.priority >= other.priority

        return False

    def __eq__(self, other):
        if isinstance(other, SeverityProperties):
            return self.priority == other.priority

        return False


vulnerability_styles = {
    "NONE": SeverityProperties(4, "green", "🟢"),
    "LOW": SeverityProperties(3, "yellow", "🟡"),
    "MEDIUM": SeverityProperties(2, Style(color=Color.from_rgb(255.0, 165.0, 0.0)), "🟠"),
    "HIGH": SeverityProperties(1, "red", "🔴"),
    "CRITICAL": SeverityProperties(0, "magenta", "🟣"),
}


class Vulnerability:
    cve_number: str
    last_modified: datetime
    severity: str
    score: float
    cvss_vector: str
    cvss_version: float
    source: str
    description: str
    cwes: list

    def __init__(self, sql_list=None, cwe_list=None):
        if sql_list is None:
            self.cve_number = ""
            self.last_modified = datetime.utcnow()
            self.severity = "NONE"
            self.score = 0
            self.cvss_vector = ""
            self.cvss_version = 0
            self.source = ""
            self.description = ""
        else:
            self.cve_number = sql_list[0]
            self.last_modified = datetime.fromisoformat(sql_list[2])
            self.severity = sql_list[3]
            self.score = float(sql_list[4])
            self.cvss_vector = sql_list[5]
            self.cvss_version = float(sql_list[6])
            self.source = sql_list[7]
            self.description = sql_list[1]

        self.last_modified = self.last_modified.replace(tzinfo=SimpleUtc())

        if cwe_list is None:
            self.cwes = []
        else:
            self.cwes = cwe_list

    def __str__(self):
        return self.rich().plain

    def rich(self):
        column_fmt_str = "{:<14}"

        def format_string(prop: str, value: str = ""):
            return Text.assemble(
                (f"{column_fmt_str.format(prop)}", "bold cyan"), ("  :  ", "yellow")
            ) + Text.from_markup(value)

        name = format_string("CVE")
        column_spacer = ("{:<" + f"{name.cell_len}" + "}").format(" ")
        rep = format_string("CVE", f"{self.generate_cve_link_text()}\n")
        rep += format_string("Last Modified", f"{self.last_modified}\n")
        rep += format_string("Severity", f"[{vulnerability_styles[self.severity].style}]{self.severity}[/]\n")
        rep += format_string("Score", f"[{vulnerability_styles[self.severity].style}]{self.score}[/]\n")
        rep += format_string("CVSS Vector", f"{self.cvss_vector}\n")
        rep += format_string("CVSS Version", f"{self.cvss_version}\n")
        rep += format_string("Source", f"{self.source}\n")
        rep += (
            format_string("Description")
            + textwrap.fill(
                self.description, width=120, initial_indent=column_spacer, subsequent_indent=column_spacer
            ).strip()
            + "\n"
        )
        if self.cwes and len(self.cwes) > 0:
            rep += format_string("CWEs")
            formatted_cwes = list()
            for cwe in self.cwes:
                formatted_cwes.append(
                    f"[link=https://cwe.mitre.org/data/definitions/{cwe.split('-')[1]}.html]{cwe}[/link]"
                )
            rep += Text.from_markup(f"\n{column_spacer}".join(formatted_cwes))

        return rep

    def __rich_repr__(self):
        return self.rich()

    def generate_cve_link_text(self):
        vulnerability_color: Color
        if isinstance(vulnerability_styles[self.severity].style, str):
            vulnerability_color = vulnerability_styles[self.severity].style
        else:
            vulnerability_color = vulnerability_styles[self.severity].style.color

        cve_link_style = Style(color=vulnerability_color)
        cve_link_style.update_link(f"https://nvd.nist.gov/vuln/detail/{self.cve_number}")
        cve_link = rich.text.Text(
            f"[link=https://nvd.nist.gov/vuln/detail/{self.cve_number}]{self.cve_number}[/link]",
            style=vulnerability_styles[self.severity].style,
        )
        cve_link.pad_right(17 - len(self.cve_number))
        return cve_link

    def __lt__(self, other):
        if isinstance(other, Vulnerability):
            if vulnerability_styles[self.severity] != vulnerability_styles[other.severity]:
                return vulnerability_styles[self.severity] < vulnerability_styles[other.severity]

            return self.cve_number < other.cve_number

        if isinstance(other, str):
            return self.cve_number < other

        return False

    def __gt__(self, other):
        if isinstance(other, Vulnerability):
            if vulnerability_styles[self.severity] != vulnerability_styles[other.severity]:
                return vulnerability_styles[self.severity] > vulnerability_styles[other.severity]

            return self.cve_number > other.cve_number

        if isinstance(other, str):
            return self.cve_number > other

        return False

    def __le__(self, other):
        if isinstance(other, Vulnerability):
            if vulnerability_styles[self.severity] != vulnerability_styles[other.severity]:
                return vulnerability_styles[self.severity] <= vulnerability_styles[other.severity]

            return self.cve_number <= other.cve_number

        if isinstance(other, str):
            return self.cve_number <= other

        return False

    def __ge__(self, other):
        if isinstance(other, Vulnerability):
            if vulnerability_styles[self.severity] != vulnerability_styles[other.severity]:
                return vulnerability_styles[self.severity] >= vulnerability_styles[other.severity]

            return self.cve_number >= other.cve_number

        if isinstance(other, str):
            return self.cve_number >= other

        return False

    def __eq__(self, other):
        if isinstance(other, Vulnerability):
            if vulnerability_styles[self.severity] != vulnerability_styles[other.severity]:
                return False

            return self.cve_number == other.cve_number

        if isinstance(other, str):
            return self.cve_number == other

        return False

    @staticmethod
    def sql_query_name_list():
        return [
            "cve_number",
            "description",
            "last_modified",
            "severity",
            "score",
            "cvss_vector",
            "cvss_version",
            "data_source",
        ]



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

import semantic_version

invalid_version_list = ["", "*", "-"]


class VersionHandler:

    def __init__(self):
        self._package = ""
        self._type = semantic_version.Version

    @property
    def package(self):
        return self._package

    @property
    def version_type(self):
        return self._type

    def convert(self, ver: str) -> str | semantic_version.Version:
        return ver



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

from rich.text import Text
from rich import print
from rich.style import Style
from rich.console import Console

import textwrap

from typing import Final

column_fmt_default_size: Final = 20
column_fmt_left_str = "{:<" + f"{column_fmt_default_size}" + "}"
column_fmt_right_str = "{:>" + f"{column_fmt_default_size}" + "}"


def format_string(prop: str, value: str = "", style: str | Style = "bold cyan", left_justify: bool = True):
    prop_fmtd_string = f"{column_fmt_left_str.format(prop)}" if left_justify else f"{column_fmt_right_str.format(prop)}"
    return Text.assemble((prop_fmtd_string, style), ("  :  ", "yellow"), f"{value}")


def print_list(name: str, data_list: list, with_wrap: bool = False):

    if not data_list or len(data_list) == 0:
        print(format_string(name, "None\n"))
        return

    name_str = format_string(name)

    column_spacer = ("{:<" + f"{name_str.cell_len}" + "}").format(" ")
    list_str = ""
    if with_wrap:
        tab_width = len(max(data_list, key=len)) + 3
        list_str = textwrap.fill(
            "\t".join(data_list),
            width=Console().width,
            initial_indent=column_spacer,
            subsequent_indent=column_spacer,
            tabsize=tab_width,
        ).strip()
    else:
        list_str = ("\n" + column_spacer).join(data_list)

    print(format_string(name) + list_str)



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

import json

from beartype.typing import Dict
from spdx_tools.spdx.model import Document as SPDXDocument
from spdx_tools.spdx.parser.jsonlikedict.json_like_dict_parser import JsonLikeDictParser

_replacement_symbols = {"&": "AND", "|": "OR"}


def parse_from_json(file):
    input_doc_as_dict: Dict = json.load(file)

    if "packages" in input_doc_as_dict.keys():
        for package in input_doc_as_dict["packages"]:
            if "licenseDeclared" in package.keys():
                new_string = package["licenseDeclared"]
                for symbol in _replacement_symbols:
                    new_string = new_string.replace(symbol, _replacement_symbols[symbol])

                package["licenseDeclared"] = new_string

    return JsonLikeDictParser().parse(input_doc_as_dict)


def parse_from_json_file(file_name: str, encoding: str = "utf-8") -> SPDXDocument:
    with open(file_name, encoding=encoding) as file:
        return parse_from_json(file)



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright 2024 Ics inc.

from .default import DefaultVersionHandler
from .openssh import OpenSSHVersionHandler
from .openssl import OpenSSLVersionHandler
from .linux_kernel import LinuxKernelVersionHandler

from ics_sbom_libs.cve_match.package_matching.versionfactory import VersionFactory

__all__ = ["default", "openssh", "openssl", "linux_kernel"]

VersionFactory.add_package_handler(DefaultVersionHandler)
VersionFactory.add_package_handler(OpenSSHVersionHandler)
VersionFactory.add_package_handler(OpenSSLVersionHandler)
VersionFactory.add_package_handler(LinuxKernelVersionHandler)



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

import re

from spdx_tools.spdx.model import Document as SPDXDocument
from spdx_tools.spdx.parser.tagvalue.parser import Parser as TagValueParser

from ics_sbom_libs.sbom_import.spdx_tag_value.filter_lexers import SimplifiedFilterLexer
from ics_sbom_libs.sbom_import.spdx_tag_value.progress_lexer import ProgressLexer


def parse_from_tag_value_file(file_name: str, encoding: str = "utf-8") -> SPDXDocument:
    parser = TagValueParser()
    parser.lex = ProgressLexer(SimplifiedFilterLexer())
    parser.lex.build(reflags=re.UNICODE)
    with open(file_name, encoding=encoding) as file:
        data = file.read()
    document: SPDXDocument = parser.parse(data)
    return document



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>


class DBProperties:
    def __init__(
        self,
        database_type: str = "",
        database_name: str = "",
        user_name: str = "",
        password: str = "",
        host_name="localhost",
        host_port="5432",
    ):
        self._connection_string: str = ""

        self._db_type = database_type  # 'postgresql'
        self._db_name = database_name  # 'cve_database'
        self._db_user = user_name  # 'postgres'
        self._db_pass = password  # 'postgres123'
        self._db_host = host_name  # 'localhost'
        self._db_port = host_port  # '5432'

        self._update_connection_string()

    @property
    def database_type(self):
        return self._db_type

    @database_type.setter
    def database_type(self, db_type: str = ""):
        if len(db_type) == 0:
            return

        self._db_type = db_type
        self._update_connection_string()

    @property
    def database_name(self):
        return self._db_name

    @database_name.setter
    def database_name(self, db_name: str = ""):
        if len(db_name) == 0:
            return

        self._db_name = db_name
        self._update_connection_string()

    @property
    def database_user(self):
        return self._db_user

    @database_user.setter
    def database_user(self, db_user: str = ""):
        if len(db_user) == 0:
            return

        self._db_user = db_user
        self._update_connection_string()

    @property
    def database_password(self):
        return self._db_pass

    @database_password.setter
    def database_password(self, db_password: str = ""):
        if len(db_password) == 0:
            return

        self._db_pass = db_password
        self._update_connection_string()

    @property
    def host_name(self):
        return self._db_host

    @host_name.setter
    def host_name(self, db_host: str = ""):
        if len(db_host) == 0:
            return

        self._db_host = db_host
        self._update_connection_string()

    @property
    def host_port(self):
        return self._db_port

    @host_port.setter
    def host_port(self, db_port: str = ""):
        if len(db_port) == 0:
            return

        self._db_port = db_port
        self._update_connection_string()

    @property
    def connection_string(self):
        self._update_connection_string()
        # Connect to the database
        return self._connection_string

    def _update_connection_string(self):
        self._connection_string = (
            f"{self._db_type}://{self._db_user}:{self._db_pass}@{self._db_host}:{self._db_port}/{self._db_name}"
        )



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

import unittest
from src.ics_sbom_libs.common.dbproperties import DBProperties


class DBPropertiesTestCase(unittest.TestCase):
    def test_raw_init(self):
        props = DBProperties()

        self.assertEqual(props.database_type, "")
        self.assertEqual(props.database_name, "")
        self.assertEqual(props.database_user, "")
        self.assertEqual(props.database_password, "")
        self.assertEqual(props.host_name, "localhost")
        self.assertEqual(props.host_port, "5432")

    def test_connection_string(self):
        props = DBProperties(
            database_type="postgresql", database_name="testdb", user_name="postgres", password="postgres123"
        )

        self.assertEqual(props.database_type, "postgresql")
        self.assertEqual(props.database_name, "testdb")
        self.assertEqual(props.database_user, "postgres")
        self.assertEqual(props.database_password, "postgres123")
        self.assertEqual(props.host_name, "localhost")
        self.assertEqual(props.host_port, "5432")

        self.assertEqual(props.connection_string, "postgresql://postgres:postgres123@localhost:5432/testdb")


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

from ics_sbom_libs.cve_match.package_matching import VersionHandler


class VersionFactory:
    _instance = None
    _handlers: dict = {}

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(VersionFactory, cls).__new__(cls)

        return cls._instance

    def _add_handler(self, handler_type):
        if not issubclass(handler_type, VersionHandler):
            return

        package_name = handler_type().package
        if package_name not in self._handlers.keys():
            self._handlers[package_name] = handler_type

    @staticmethod
    def get_handler(package_name: str) -> VersionHandler:
        factory = VersionFactory()
        return (
            factory._handlers[package_name]()
            if package_name in factory._handlers.keys()
            else factory._handlers["default"]()
        )

    @staticmethod
    def add_package_handler(handler_type):
        factory = VersionFactory()
        factory._add_handler(handler_type)



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

import argparse

__all__ = ["setup_log_arg", "handle_log_arg"]

_logList = ["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
_log_default_level = "WARNING"


def setup_log_arg(parser: argparse.ArgumentParser):
    if not parser:
        return

    parser.add_argument("--log", type=str, default=_log_default_level, help="Set The Log level of the application")


def handle_log_arg(args: argparse.Namespace):
    if not args or args.log is None:
        return _log_default_level

    log_value = args.log.upper()
    return log_value if log_value in _logList else _log_default_level



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

import semantic_version

from ics_sbom_libs.cve_match.package_matching import VersionHandler, invalid_version_list


class OpenSSHVersionHandler(VersionHandler):

    def __init__(self):
        super(OpenSSHVersionHandler, self).__init__()
        self._package = "openssh"
        self._type = semantic_version.Version

    def convert(self, ver: str) -> str | semantic_version.Version:
        if ver in invalid_version_list:
            return ver

        if ver.find("p") != -1:
            ver = ver[:-2]

        try:
            pkg_version = self.version_type.coerce(ver)
            return pkg_version

        except ValueError:
            return ver



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

from ply.lex import TOKEN
from spdx_tools.spdx.parser.tagvalue.lexer import SPDXLexer


class SimplifiedFilterLexer(SPDXLexer):
    @TOKEN(r"\s*Relationship:.+")
    def t_RELATIONSHIP(self, t):
        pass

    @TOKEN(r"\s*RelationshipComment:.+")
    def t_RELATIONSHIP_COMMENT(self, t):
        pass

    @TOKEN(r"\s*FileName:.+")
    def t_FILE_NAME(self, t):
        pass

    @TOKEN(r"\s*FileType:.+")
    def t_FILE_TYPE(self, t):
        pass

    @TOKEN(r"\s*FileChecksum:.+")
    def t_FILE_CHECKSUM(self, t):
        pass

    @TOKEN(r"\s*LicenseConcluded:.+")
    def t_FILE_LICENSE_CONCLUDED(self, t):
        pass

    @TOKEN(r"\s*LicenseInfoInFile:.+")
    def t_FILE_LICENSE_INFO(self, t):
        pass

    @TOKEN(r"\s*FileCopyrightText:.+")
    def t_FILE_COPYRIGHT_TEXT(self, t):
        pass

    @TOKEN(r"\s*LicenseComments:.+")
    def t_FILE_LICENSE_COMMENT(self, t):
        pass

    @TOKEN(r"\s*FileComment:.+")
    def t_FILE_COMMENT(self, t):
        pass

    @TOKEN(r"\s*FileNotice:.+")
    def t_FILE_NOTICE(self, t):
        pass

    @TOKEN(r"\s*FileContributor:.+")
    def t_FILE_CONTRIBUTOR(self, t):
        pass

    @TOKEN(r"\s*FileAttributionText:.+")
    def t_FILE_ATTRIBUTION_TEXT(self, t):
        pass



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

import semantic_version


class BuildVersion(semantic_version.Version):

    def __init__(
        self, version_string=None, major=None, minor=None, patch=None, prerelease=None, build=None, partial=False
    ):
        super(BuildVersion, self).__init__(version_string, major, minor, patch, prerelease, build, partial)

        # _cmp_precedence_key is used for semver-precedence comparison
        self._cmp_precedence_key = self._build_precedence_key(with_build=True)



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

import semantic_version

from ics_sbom_libs.cve_match.package_matching import VersionHandler, invalid_version_list


class DefaultVersionHandler(VersionHandler):

    def __init__(self):
        super(DefaultVersionHandler, self).__init__()
        self._package = "default"
        self._type = semantic_version.Version

    def convert(self, ver: str) -> str | semantic_version.Version:
        try:
            pkg_version = self.version_type.coerce(ver) if ver not in invalid_version_list else ver
        except ValueError:
            pkg_version = ver

        return pkg_version



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Chris Rizzitello <crizzitello@ics.com>
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>
# SPDX-FileContributor: Milo Kerr <mkerr@ics.com>

import datetime
import os

from enum import Enum
from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor, as_completed

from .matchresult import MatchResult
import semantic_version
import pathlib

from rich import table, print

from beartype.typing import Optional
from cpeparser import CpeParser
from spdx_tools.spdx.model import Document as SPDXDocument
from spdx_tools.spdx.model import CreationInfo as SPDXCreationInfo
from spdx_tools.spdx.model import Package as SPDXPackage
from spdx_tools.spdx.model import Actor as SPDXActor
from spdx_tools.spdx.model import SpdxNoAssertion
from spdx_tools.spdx.model import ActorType as SPDXActorType
from spdx_tools.spdx.model import ExternalPackageRef as SPDXExternalPackageRef
from spdx_tools.spdx.model import ExternalPackageRefCategory as SPDXExternalRefCategory

from ics_sbom_libs.cve_match.package_matching.versionfactory import VersionFactory
from ics_sbom_libs.cve_match.cpe_match_results import CpeMatchResult

from ics_sbom_libs.common.vulnerability import vulnerability_styles
from ics_sbom_libs.cve_fetch.vulnerabilitydatabase import VulnerabilityDatabase


class MatchTableOutput(Enum):
    CvesOnly = 1
    WithoutCvesOnly = 2
    All = 3


def cpe_factory(product: str, version: Optional[str] = None, vendor: Optional[str] = None):
    _vendor = vendor if vendor else "*"
    _version = version if version else "*"

    return f"cpe:2.3:a:{_vendor}:{product}:{_version}:*:*:*:*:*:*:*"


class CveMatcher:
    db_path: pathlib.Path
    _spdx_document: SPDXDocument

    # Results
    result_list: list[MatchResult]

    # Results meta data
    total_package_count: int
    dirty_package_count: int
    clean_package_count: int
    total_cve_count: int

    # Timing meta data
    scanTime: str

    def __init__(self, db_path: pathlib.Path):
        self.db_path = db_path

        self.total_package_count = 0
        self.dirty_package_count = 0
        self.clean_package_count = 0
        self.total_cve_count = 0

        creation_info = SPDXCreationInfo(
            spdx_version="SPDX-2.3",
            spdx_id="SPDXRef-DOCUMENT",
            name="ICS SBOM Package Search",
            document_namespace="",
            creators=[SPDXActor(SPDXActorType.TOOL, "icsbom")],
            created=datetime.datetime.now(datetime.timezone.utc),
        )
        self._spdx_document = SPDXDocument(creation_info=creation_info)

    @property
    def spdx_document(self):
        return self._spdx_document

    @spdx_document.setter
    def spdx_document(self, document):
        if not isinstance(document, SPDXDocument):
            print("[yellow][b]WARNING:[/b][/yellow] The provided document is not an SPDXDocument.")
            return

        self._spdx_document = document

    def add_package(self, package: str, version: Optional[str] = None, vendor: Optional[str] = None):
        if len(package) == 0:
            print("[red][b]ERROR:[/b] The package name is empty. Not adding to Matcher.[/red]")
            return

        new_package = SPDXPackage(
            name=package,
            spdx_id=f"SPDXRef-{package}",
            version=version,
            supplier=SpdxNoAssertion(),
            download_location=SpdxNoAssertion(),
            description="No Text",
        )

        if vendor:
            new_package.external_references = [
                SPDXExternalPackageRef(
                    SPDXExternalRefCategory.SECURITY,
                    "http://spdx.org/rdf/references/cpe23Type",
                    cpe_factory(package, version, vendor),
                )
            ]

        self._spdx_document.packages.append(new_package)

    def process(self):
        if len(self.spdx_document.packages) == 0:
            print("[red][b]ERROR:[/b] No SPDX Document has no packages[/red]")
            raise RuntimeError("No SPDX Document has no packages")

        self.total_package_count = len(self.spdx_document.packages)
        self.dirty_package_count = 0
        self.clean_package_count = 0
        self.total_cve_count = 0

        self.result_list = process(self.spdx_document, self.db_path)
        self.scanTime = str(datetime.datetime.now(datetime.timezone.utc)).replace(" ", "T")[:-7] + "Z"

        for result in self.result_list:
            if result.cve_list:
                self.dirty_package_count += 1
                self.total_cve_count += len(result.cve_list)
            else:
                self.clean_package_count += 1

    def create_match_table(self, table_output: MatchTableOutput = MatchTableOutput.All):
        match_table = table.Table(title="CVE Results", row_styles=["dim", ""], expand=True)
        match_table.add_column(header="Package", style="green")
        match_table.add_column(header="Version", style="magenta")
        match_table.add_column(header="CVE Count", style="blue")
        match_table.add_column(header="None", style=vulnerability_styles["NONE"].style)
        match_table.add_column(header="Low", style=vulnerability_styles["LOW"].style)
        match_table.add_column(header="Medium", style=vulnerability_styles["MEDIUM"].style)
        match_table.add_column(header="High", style=vulnerability_styles["HIGH"].style)
        match_table.add_column(header="Critical", style=vulnerability_styles["CRITICAL"].style)
        match_table.add_column(header="CVEs")

        vuln_info_counts = {"Total": 0, "NONE": 0, "LOW": 0, "MEDIUM": 0, "HIGH": 0, "CRITICAL": 0}
        for result in self.result_list:
            vuln_info = result.get_severity_info()
            if result.cve_list and table_output != MatchTableOutput.WithoutCvesOnly:
                sorted_cves = sorted(result.cve_list)
                formated_cves = " ".join(
                    f"{vulnerability_styles[cve.severity].indicator}{cve.generate_cve_link_text()}"
                    for cve in sorted_cves
                )
                match_table.add_row(
                    f"{result.name}",
                    f"{result.version}",
                    f"{len(result.cve_list)}",
                    f'{vuln_info["NONE"]}',
                    f'{vuln_info["LOW"]}',
                    f'{vuln_info["MEDIUM"]}',
                    f'{vuln_info["HIGH"]}',
                    f'{vuln_info["CRITICAL"]}',
                    f"{formated_cves}",
                )

                vuln_info_counts["Total"] += len(result.cve_list)
                for severity in list(vuln_info_counts.keys())[1:]:
                    vuln_info_counts[severity] += vuln_info[severity]

            elif not result.cve_list and table_output != MatchTableOutput.CvesOnly:
                match_table.add_row(
                    f"{result.name}",
                    f"{result.version}",
                    f"{len(result.cve_list)}",
                    f'{vuln_info["NONE"]}',
                    f'{vuln_info["LOW"]}',
                    f'{vuln_info["MEDIUM"]}',
                    f'{vuln_info["HIGH"]}',
                    f'{vuln_info["CRITICAL"]}',
                    "",
                )

        match_table.add_section()
        match_table.add_row(
            "Totals",
            "",
            f'{vuln_info_counts["Total"]}',
            f'{vuln_info_counts["NONE"]}',
            f'{vuln_info_counts["LOW"]}',
            f'{vuln_info_counts["MEDIUM"]}',
            f'{vuln_info_counts["HIGH"]}',
            f'{vuln_info_counts["CRITICAL"]}',
            "",
        )
        return match_table

    def get_severity_info(self):
        vuln_info_counts = {"Total": 0, "NONE": 0, "LOW": 0, "MEDIUM": 0, "HIGH": 0, "CRITICAL": 0}
        for result in self.result_list:
            vuln_info = result.get_severity_info()
            vuln_info_counts["Total"] += len(result.cve_list)
            for severity in list(vuln_info_counts.keys())[1:]:
                vuln_info_counts[severity] += vuln_info[severity]

        return vuln_info_counts

    def __str__(self):
        return str(self.create_match_table())


def process(spdx_document: SPDXDocument, db_path: pathlib.Path):
    results_list = []
    unique_cpes = {}
    use_parallel = os.environ["MATCH_USE_PARALLEL"].upper() == "TRUE" if "MATCH_USE_PARALLEL" in os.environ else True
    match_results = {}

    package_pbar = tqdm(
        total=len(spdx_document.packages), desc="Matching CPEs", unit="packages", mininterval=0, miniters=1
    )
    if use_parallel:
        with ProcessPoolExecutor() as executor:
            future_to_package = {
                executor.submit(process_spdx_package, package, db_path): package for package in spdx_document.packages
            }
            for future in as_completed(future_to_package):
                package = future_to_package[future]
                result, unique_cpes_partial = future.result()
                match_results.update(result)
                for cpe, packages in unique_cpes_partial.items():
                    if cpe not in unique_cpes:
                        unique_cpes[cpe] = packages
                    else:
                        unique_cpes[cpe].extend(packages)
                package_pbar.update()
    else:
        for package in spdx_document.packages:
            result, unique_cpes_partial = process_spdx_package(package, db_path)
            match_results.update(result)
            for cpe, packages in unique_cpes_partial.items():
                if cpe not in unique_cpes:
                    unique_cpes[cpe] = packages
                else:
                    unique_cpes[cpe].extend(packages)
            package_pbar.update()

    pbar = tqdm(
        total=len(unique_cpes),
        desc="Checking CPEs for Known Issues",
        unit="cpes",
        mininterval=0,
        miniters=1,
        position=0,
        leave=True,
    )
    if use_parallel:
        with ProcessPoolExecutor() as executor:
            future_result = {executor.submit(find_cves_for_cpe, cpe, db_path): cpe for cpe in unique_cpes}

            for result in as_completed(future_result):
                results_list.append(result.result())
                pbar.update()

    else:
        for cpe in unique_cpes:
            result = find_cves_for_cpe(cpe, db_path)
            results_list.append(result)
            pbar.update()

    for cpe in results_list:
        if not cpe.cve_list:
            continue

        for package in unique_cpes[cpe.cpe]:
            match_results[package.name].cve_list += cpe.cve_list
            break

    return list(match_results.values())


def lookup_cpe_for_package(package_name: str, db_path: pathlib.Path) -> list[str] | None:
    db = VulnerabilityDatabase(db_path.parent, db_path.name)
    cpe_strings = []

    query = f"SELECT cpe FROM cpe_dictionary WHERE product='{package_name}' AND deprecated='0'"

    cursor = db.query_cache(query)
    if not cursor:
        return cpe_strings

    res = cursor.fetchall()

    if not res:
        return cpe_strings

    for cpe in res:
        cpe_strings.append(cpe[0])

    return cpe_strings


def create_cpe_string(cpe: dict) -> str:
    return CpeParser().format_prefix + ":".join(cpe.values())


def find_cves_for_cpe(cpe: str, db_path: pathlib.Path) -> CpeMatchResult:
    db = VulnerabilityDatabase(db_path.parent, db_path.name)
    # db = sqlite3.connect(db_path)

    result = find_cve_with_cpe(cpe, db)

    return result


def find_cve_with_cpe(cpe: str, db: VulnerabilityDatabase) -> CpeMatchResult:
    result = CpeMatchResult(cpe)
    vendor, product, version = result.get_cpe_properties()

    query = f"SELECT DISTINCT cve_number FROM cve_range WHERE product='{product}'"
    second_query = f"AND product='{product}'"
    if vendor != "*":
        query += f" AND vendor='{vendor}'"
        second_query += f" AND vendor='{vendor}'"

    cursor = db.query_cache(query)
    if not cursor:
        return result

    res = cursor.fetchall()

    try:
        if not res:
            return result
        for cve in res:
            include = cve_version_included(db, cve[0], product, version, sql_ex=second_query)

            if include:
                vuln = db.get_cve(cve[0])
                result.append_cve(vuln)

    except ValueError as vError:
        print(
            f"[red][b]ERROR:[/b] While processing {product} by {vendor} with version {version} had error:"
            f" {vError}[/red]"
        )

    return result


def cve_version_included(db: VulnerabilityDatabase, cve_id, package_name, package_version_str, sql_ex: str):
    cve = cve_id
    package_version = VersionFactory.get_handler(package_name).convert(package_version_str)
    if not sql_ex:
        sql_ex = f"AND product LIKE '{package_name}'"
    query = (
        "SELECT DISTINCT vulnerable, version, versionStartIncluding, versionStartExcluding, versionEndIncluding,"
        f" versionEndExcluding FROM cve_range WHERE cve_number='{cve}' {sql_ex}"
    )
    cursor = db.query_cache(query)

    results = cursor.fetchall()
    for version in results:
        vulnerable = bool(version[0])

        if not vulnerable:
            # The simplest way to deal with this is to ignore the result if it's not a true vulnerability.
            continue

        handler = VersionFactory.get_handler(package_name)
        version_exact = handler.convert(version[1])
        version_starting_including = version[2]
        version_starting_excluding = version[3]
        version_end_including = version[4]
        version_end_excluding = version[5]
        if version_exact == "*" and isinstance(package_version, semantic_version.Version):
            version_start = handler.version_type.coerce("0")
            inc_low = True
            if version_starting_including:
                version_start = handler.convert(version_starting_including)
            elif version_starting_excluding:
                version_start = handler.convert(version_starting_excluding)
                inc_low = False

            version_end = handler.version_type.coerce("10000")
            inc_high = True
            if version_end_including:
                version_end = handler.convert(version_end_including)
            elif version_end_excluding:
                version_end = handler.convert(version_end_excluding)
                inc_high = False

            if isinstance(version_start, str) or isinstance(version_end, str):
                return True
            if inc_low and inc_high and (package_version >= version_start) and (package_version <= version_end):
                return True
            if inc_low and (package_version >= version_start) and (package_version < version_end):
                return True
            if inc_high and (package_version > version_start) and (package_version <= version_end):
                return True
            if (package_version > version_start) and (package_version < version_end):
                return True

        if isinstance(package_version, str) and package_version == "-":
            return True

        if package_version == version_exact:
            return True
    return False


def generate_cpe_list(ext_refs) -> list[str]:
    cpe_list: list[str] = []
    for ref in ext_refs:
        if ref.category is SPDXExternalRefCategory.SECURITY and ref.reference_type.find("/cpe") != -1:
            cpe = ref.locator
            cpe_list.append(cpe)
    return cpe_list


def process_spdx_package(spdx_package, db_path):
    match = MatchResult(name=spdx_package.name, version=spdx_package.version)
    unique_cpes = {}
    match_results = {}

    cpes = generate_cpe_list(spdx_package.external_references)
    if cpes:
        for cpe in cpes:
            if cpe not in unique_cpes.keys():
                unique_cpes[cpe] = [spdx_package]
            else:
                unique_cpes[cpe].append(spdx_package)
            match.cpe_list.append(cpe)
    else:
        looked_up_cpes = lookup_cpe_for_package(spdx_package.name, db_path)
        new_cpe = CpeParser().parser(cpe_factory(spdx_package.name, spdx_package.version))
        if not looked_up_cpes:
            new_cpe_str = create_cpe_string(new_cpe)
            if new_cpe_str not in unique_cpes:
                unique_cpes[new_cpe_str] = [spdx_package]
            match.cpe_list.append(new_cpe_str)
        else:
            for cpe in looked_up_cpes:
                parsed_cpe = CpeParser().parser(cpe)
                new_cpe["part"] = parsed_cpe["part"]
                new_cpe["vendor"] = parsed_cpe["vendor"]
                new_cpe_str = create_cpe_string(new_cpe)
                if new_cpe_str not in unique_cpes.keys():
                    unique_cpes[new_cpe_str] = [spdx_package]
                else:
                    unique_cpes[new_cpe_str].append(spdx_package)
                if new_cpe_str not in match.cpe_list:
                    match.cpe_list.append(new_cpe_str)

    match_results[spdx_package.name] = match
    return match_results, unique_cpes



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>
# SPDX-FileContributor: Chris Rizzitello <crizzitello@ics.com>

import copy
import logging
import re
import json

import pathlib
import sys

if sys.version_info >= (3, 14):
    import tarfile
else:
    from backports.zstd import tarfile
import argparse

from cpeparser import CpeParser
from tqdm import tqdm
from datetime import datetime
from spdx_tools.spdx.formats import file_name_to_format, FileFormat
from spdx_tools.spdx.parser.parse_anything import parse_file as spdx_parse_file
from spdx_tools.spdx.parser.error import SPDXParsingError

from spdx_tools.spdx.model import Document as SPDXDocument, SpdxNone, ExternalPackageRef, ExternalPackageRefCategory
from spdx_tools.spdx.model import CreationInfo as SPDXCreationInfo
from spdx_tools.spdx.model import Actor as SPDXActor
from spdx_tools.spdx.model import ActorType as SPDXActorType

from ics_sbom_libs.sbom_import.spdx_tag_value.parse import parse_from_tag_value_file
from ics_sbom_libs.sbom_import.spdx_json.parse import parse_from_json_file, parse_from_json

log = logging.getLogger(__name__)


class FilterList:
    def __init__(self):

        self._pkg_substitution = {
            "qtbase": {"duplicate": {"rename": "qt"}, "sub_cpe": {"product": {"orig": "qt", "new": "<name>"}}},
            "qtsvg": {"sub_cpe": {"product": {"orig": "qt", "new": "<name>"}}},
            "qtdeclarative": {"sub_cpe": {"product": {"orig": "qt", "new": "<name>"}}},
            "qtgraphicaleffects": {"sub_cpe": {"product": {"orig": "qt", "new": "<name>"}}},
            "qtmultimedia": {"sub_cpe": {"product": {"orig": "qt", "new": "<name>"}}},
            "qtquickcontrols": {"sub_cpe": {"product": {"orig": "qt", "new": "<name>"}}},
            "qtquickcontrols2": {"sub_cpe": {"product": {"orig": "qt", "new": "<name>"}}},
            "qtserialport": {"sub_cpe": {"product": {"orig": "qt", "new": "<name>"}}},
            "qttools": {"sub_cpe": {"product": {"orig": "qt", "new": "<name>"}}},
            "qtvirtualkeyboard": {"sub_cpe": {"product": {"orig": "qt", "new": "<name>"}}},
            "qtwebsockets": {"sub_cpe": {"product": {"orig": "qt", "new": "<name>"}}},
            "qtxmlpatterns": {"sub_cpe": {"product": {"orig": "qt", "new": "<name>"}}},
            "flex": {"add_cpe": {"vendor": "westes", "product": "<name>", "version": "<version>"}},
            "dbus": {"duplicate": {"rename": "libdbus", "sub_cpe": {"product": {"orig": "dbus", "new": "<name>"}}}},
            "flac": {
                "sub_cpe": {
                    "vendor": {"orig": "*", "new": "flac_project"},
                    "product": {"orig": "libflac", "new": "<name>"},
                }
            },
            "bzip2": {"add_cpe": {"vendor": "bzip", "product": "compress-raw-bzip2", "version": "<version>"}},
            "libflac++6": {"add_cpe": {"vendor": "flac_project", "product": "flac", "version": "<version>"}},
            "libflac8": {"add_cpe": {"vendor": "flac_project", "product": "flac", "version": "<version>"}},
            "curl": {
                "duplicate": {"rename": "libcurl", "rem_cpe": {"product": "curl"}},
                "rem_cpe": {"product": "libcurl"},
            },
            "libcurl3": {"rename": "libcurl"},
            "libcurl4": {"rename": "libcurl"},
            "expat": {
                "duplicate": {"rename": "libexpat", "rem_cpe": {"product": "expat"}},
                "rem_cpe": {"product": "libexpat"},
            },
            "file": {"add_cpe": {"vendor": "file_project", "product": "<name>", "version": "<version>"}},
            "perl": {"add_cpe": {"vendor": "perl", "product": "<name>", "version": "<version>"}},
        }

        self._exclusion_list = [
            "-doc",
            "-dev",
            "-dbg",
            "locale",
            "-ptest",
            "-tests",
            "-examples",
            "-mkspecs",
            "-staticdev",
            "qttranslations",
            "packagegroup",
            "glibc-gconv",
            "glibc-charmap",
            "tzdata-",
            "-completion",
            "-data",
            "-src",
            "-native",
            "-cross-",
            "-source-",
            "-headers",
            "index.json",
        ]

    @property
    def exclusions(self):
        return self._exclusion_list

    @exclusions.setter
    def exclusions(self, new_list: list):
        if not new_list:
            return

        # removes any duplicates.
        self._exclusion_list = list(set(new_list))

    @property
    def substitutions(self):
        return self._pkg_substitution

    @substitutions.setter
    def substitutions(self, new_substitutions: dict):
        if not new_substitutions:
            return

        self._pkg_substitution = new_substitutions

    def add_exclusion(self, exclude: str):
        if exclude not in self._exclusion_list:
            self._exclusion_list.append(exclude)

    def remove_exclusion(self, remove: str):
        if remove in self._exclusion_list:
            self._exclusion_list.remove(remove)

    def set_filters_from_file(self, list_file_path: pathlib.Path):
        if not list_file_path or not list_file_path.exists():
            return

        with open(list_file_path, "r") as new_list_file:
            filters = json.load(new_list_file)

            if "exclusions" in filters.keys():
                self.exclusions = filters["exclusions"]

            if "substitutions" in filters.keys():
                self.substitutions = filters["substitutions"]

    def compile_exclusions(self):
        return re.compile("|".join(self._exclusion_list))

    def compile_substitutions(self):
        return re.compile("|".join(self._pkg_substitution.keys()))


class FilteredParser:

    def __init__(self):
        self._filter = FilterList()
        self._encoding = "utf-8"
        self._tar_dir_pattern = ""
        self._regex_pattern = ""

    @property
    def encoding(self):
        return self._encoding

    @encoding.setter
    def encoding(self, value: str):
        if not value:
            return

        if value != "utf-8":
            log.warning(
                "It's recommended to use the UTF-8 encoding for any SPDX file. "
                "Consider changing the encoding of the file."
            )

        self._encoding = value

    @property
    def filter_list(self):
        return self._filter

    @filter_list.setter
    def filter_list(self, new_list: FilterList):
        if not new_list:
            return

        self._filter = new_list

    @property
    def tar_dir_pattern(self):
        return self._tar_dir_pattern

    @tar_dir_pattern.setter
    def tar_dir_pattern(self, new_pattern: str):
        if not new_pattern:
            return

        self._tar_dir_pattern = new_pattern

    @property
    def regex_pattern(self):
        return self._regex_pattern

    @regex_pattern.setter
    def regex_pattern(self, new_pattern: str):
        if not new_pattern:
            return
        try:
            self._regex_pattern = re.compile(re.escape(new_pattern))
        except re.error:
            logging.warning("Failed to parse regular expression")

    @staticmethod
    def setup_args(parser: argparse.ArgumentParser):
        if not parser:
            return

        parser.add_argument(
            "--filter_file",
            type=pathlib.Path,
            default="",
            help="Sets the up the filtering for substitutions and exclusions to the " "contents of the file",
        )

        parser.add_argument(
            "--write_filters",
            type=pathlib.Path,
            default=None,
            help="Writes a filter file that contains the default filters",
        )

        parser.add_argument(
            "-t",
            "--tar_dir_pattern",
            type=str,
            help="sets a dir name pattern to search for in the tar file if a tar file is given "
            'as the input file. ["recipes"|"packages"]',
        )

        parser.add_argument(
            "-r",
            "--regex_pattern",
            type=str,
            help="sets a name pattern to search for in the tar file if a tar file is given "
            r"as the input file. [\/recipe-\S+$|^recipe-\S+$)]",
        )

    def process_args(self, args):
        if not args:
            return

        if args.tar_dir_pattern and args.regex_pattern:
            logging.warning("regex_pattern replaces the tar_dir_pattern argument, tar_dir_pattern will not be used")
            args.tar_dir_pattern = ""

        if args.filter_file.is_file():
            self.filter_list.set_filters_from_file(args.filter_file)

        if args.tar_dir_pattern:
            if args.tar_dir_pattern == "recipes" and not args.regex_pattern:
                args.regex_pattern = r"(\/recipe-\S+$|^recipe-\S+$)"
            else:
                self._tar_dir_pattern = args.tar_dir_pattern

        if args.regex_pattern:
            self._regex_pattern = args.regex_pattern

        if args.write_filters is not None:
            f = open(args.write_filters, "w")
            fout = '{\n  "substitutions": %s,LAST_SUB\n  "exclusions": %sLAST_EXCLUDE\n}' % (
                json.dumps(self.filter_list.substitutions, indent=4),
                json.dumps(self.filter_list.exclusions, indent=4),
            )
            fout = fout.replace("},LAST_SUB", "  },")
            fout = fout.replace("]LAST_EXCLUDE", "  ]")
            f.write(fout)
            f.close()

    def parse(self, sbom_name: pathlib.Path):
        doc: SPDXDocument | None = None

        if sbom_name.is_file() and tarfile.is_tarfile(sbom_name):
            doc = self._parse_tar(sbom_name)

        elif sbom_name.is_file():
            doc = self._parse_file(sbom_name)

        elif sbom_name.is_dir():
            doc = self._parse_dir(sbom_name)

        if not doc:
            return None

        def handle_substitutions(sub_package, package_substitutions):
            return_packages = [sub_package]
            for sub_command in package_substitutions.keys():
                if sub_command == "remove":
                    return []

                if sub_command == "rename":
                    sub_package.name = package_substitutions[sub_command]

                elif sub_command == "duplicate":
                    new_package = copy.deepcopy(sub_package)
                    return_packages += handle_substitutions(new_package, package_substitutions[sub_command])

                elif sub_command == "sub_cpe":
                    for ref in sub_package.external_references:
                        if (
                            ref.category is ExternalPackageRefCategory.SECURITY
                            and ref.reference_type.find("/cpe") != -1
                        ):
                            cpe = CpeParser().parser(ref.locator)
                            for part in package_substitutions[sub_command]:
                                if part == "version" and package_substitutions[sub_command][part]["orig"] == "*":
                                    if package_substitutions[sub_command][part]["new"] == "<version>":
                                        cpe[part] = sub_package.version
                                    else:
                                        cpe[part] = package_substitutions[sub_command][part]["new"]
                                if cpe[part] == package_substitutions[sub_command][part]["orig"]:
                                    if package_substitutions[sub_command][part]["new"] == "<name>":
                                        cpe[part] = sub_package.name
                                    else:
                                        cpe[part] = package_substitutions[sub_command][part]["new"]
                            ref.locator = CpeParser().format_prefix + ":".join(cpe.values())

                elif sub_command == "rem_cpe":
                    remove_refs = []
                    for ref in sub_package.external_references:
                        if (
                            ref.category is ExternalPackageRefCategory.SECURITY
                            and ref.reference_type.find("/cpe") != -1
                        ):
                            cpe = CpeParser().parser(ref.locator)
                            for part in package_substitutions[sub_command]:
                                if cpe[part] == package_substitutions[sub_command][part]:
                                    remove_refs.append(ref)

                    for ref in remove_refs:
                        sub_package.external_references.remove(ref)

                elif sub_command == "add_cpe":
                    new_cpe = CpeParser().parser("cpe:2.3:a:*:*:*:*:*:*:*:*:*:*")
                    for part in package_substitutions[sub_command]:
                        if package_substitutions[sub_command][part] == "<version>":
                            new_cpe[part] = sub_package.version
                        elif package_substitutions[sub_command][part] == "<name>":
                            new_cpe[part] = sub_package.name
                        else:
                            new_cpe[part] = package_substitutions[sub_command][part]
                    ref = ExternalPackageRef(
                        category=ExternalPackageRefCategory.SECURITY,
                        reference_type="http://spdx.org/rdf/references/cpe23Type",
                        locator=CpeParser().format_prefix + ":".join(new_cpe.values()),
                    )
                    sub_package.external_references.append(ref)

            return return_packages

        exclusions = self._filter.compile_exclusions()
        substitutions = list(self._filter.substitutions.keys())

        packages = sorted(doc.packages, key=lambda pkg: pkg.name)
        doc.packages.clear()
        for package in packages:
            added_packages = []
            if exclusions.search(package.name):
                continue

            if package.name in substitutions:
                added_packages = handle_substitutions(package, self._filter.substitutions[package.name])
            else:
                added_packages = [package]

            doc.packages += added_packages

        doc.packages = sorted(doc.packages, key=lambda pkg: pkg.name)
        return doc

    def _parse_file(self, sbom_file_name: pathlib.Path):
        input_format = file_name_to_format(str(sbom_file_name))

        try:
            if input_format == FileFormat.TAG_VALUE:
                return parse_from_tag_value_file(str(sbom_file_name), self.encoding)
            elif input_format == FileFormat.JSON:
                return parse_from_json_file(str(sbom_file_name), self.encoding)
            else:
                # Use the more general parser
                return spdx_parse_file(str(sbom_file_name), self.encoding)
        except SPDXParsingError as e:
            messages = e.messages
            logging.warning(f"Error processing {sbom_file_name.name}: {messages[0]}")

        return None

    def _parse_dir(self, sbom_dir_name: pathlib.Path):
        doc: SPDXDocument | None = None

        if not sbom_dir_name or not sbom_dir_name.is_dir():
            logging.warning(f"Given path, {sbom_dir_name}, does not exist or is not a directory.")
            return doc

        actor = SPDXActor(actor_type=SPDXActorType.ORGANIZATION, name="ics")
        doc_ci = SPDXCreationInfo(
            spdx_version="SPDX-2.3",
            spdx_id="SPDXRef_DOCUMENT",
            name="no-name",
            document_namespace="com.ics.no-name",
            created=datetime.now(),
            creators=[actor],
        )
        doc = SPDXDocument(doc_ci)
        name_filter = self._filter.compile_exclusions()

        files = [f for f in sbom_dir_name.iterdir()]
        for file in (pbar := tqdm(files, desc="Processing packages", unit="package")):
            if name_filter.search(file.name):
                continue

            pbar.set_postfix_str(str(file.name))
            file_doc = self._parse_file(file)
            doc.packages += file_doc.packages
            doc.files += file_doc.files
            doc.snippets += file_doc.snippets
            doc.annotations += file_doc.annotations
            doc.relationships += file_doc.relationships
            doc.extracted_licensing_info += file_doc.extracted_licensing_info

        return doc

    def _parse_tar(self, sbom_tar_name: pathlib.Path):
        doc: SPDXDocument | None = None

        if not sbom_tar_name or not sbom_tar_name.is_file():
            logging.warning(f"Given path, {sbom_tar_name}, does not exist.")
            return doc

        try:
            tf = tarfile.open(sbom_tar_name)
        except tarfile.ReadError:
            logging.warning(f"Failed to read, {sbom_tar_name}")
            return doc

        actor = SPDXActor(actor_type=SPDXActorType.ORGANIZATION, name="ics")
        doc_ci = SPDXCreationInfo(
            spdx_version="SPDX-2.3",
            spdx_id="SPDXRef_DOCUMENT",
            name="no-name",
            document_namespace="com.ics.no-name",
            created=datetime.now(),
            creators=[actor],
        )
        doc = SPDXDocument(doc_ci)
        name_filter = self._filter.compile_exclusions()

        pattern = []

        if self._regex_pattern:
            fileList = [member.name for member in tf.getmembers()]
            files = "\n".join(fileList)
            pattern = re.findall(self._regex_pattern, files, re.MULTILINE)
        elif self._tar_dir_pattern:
            pattern = [
                member.name for member in tf.getmembers() if member.isdir() and self._tar_dir_pattern in member.name
            ]

        files: list

        if pattern:
            files = [f for p in pattern for f in tf.getmembers() if p in f.name and not f.isdir()]
        else:
            files = [f for f in tf.getmembers() if not f.isdir() and ".spdx" in f.name]

        for file in (pbar := tqdm(files, desc="Processing packages", unit="package")):
            if name_filter.search(file.name):
                continue

            pname = pathlib.Path(file.name)
            if pname.name.startswith("."):
                continue

            pbar.set_postfix_str(pname.name)
            contents = tf.extractfile(file)
            file_doc = parse_from_json(contents)
            doc.packages += file_doc.packages
            doc.files += file_doc.files
            doc.snippets += file_doc.snippets
            doc.annotations += file_doc.annotations
            doc.relationships += file_doc.relationships
            doc.extracted_licensing_info += file_doc.extracted_licensing_info

        return doc


def parse_anything(sbom_name: pathlib.Path, encoding: str = "utf-8"):
    """
    Uses the default FilterParser to process either a SBoM file or a directory of SBoM files.

    :param sbom_name:  The name of the SBoM file or SBoM files directory
    :param encoding: Text encoding of the file. (Default: "utf-8")
    :return: SPDX Document
    """
    if encoding != "utf-8":
        logging.warning(
            "It's recommended to use the UTF-8 encoding for any SPDX file. Consider changing the encoding of the file."
        )

    if sbom_name == pathlib.Path() or not (sbom_name.is_dir() or sbom_name.is_file()):
        logging.warning(f"Given path, {sbom_name}, does not exist or is not a directory.")
        return None

    parser = FilteredParser()
    parser.encoding = encoding

    return parser.parse(sbom_name)


def print_package_table(doc: SPDXDocument):
    """
    Prints a table of the packages from the given doc

    :param doc:  The SBoM document
    """
    name_width = 30
    version_width = 30
    license_width = 20
    vendors_width = 55
    related_width = 55

    # Get the max width of each column
    for package in doc.packages:
        package_license = (
            package.license_declared.get_literals()[0].key
            if package.license_declared is not None and not isinstance(package.license_declared, SpdxNone)
            else ""
        )
        name_width = name_width if len(package.name) < name_width else len(package.name)
        version_width = version_width if len(package.version) < version_width else len(package.version)
        license_width = license_width if len(package_license) < license_width else len(package_license)

    # Set the format
    table_format = " ".join(
        [
            "\t{:>5}",
            "{:<" + f"{name_width}" + "}",
            "{:<" + f"{version_width}" + "}",
            "{:<" + f"{license_width}" + "}",
            "{:<" + f"{vendors_width}" + "}",
            "{:<" + f"{related_width}" + "}",
        ]
    )

    # Print Header
    print("Packages Table:")
    print(table_format.format("     ", "Package Name", "Version", "License", "Vendors", "Related"))

    # Print Table
    package_index = 1
    for package in doc.packages:
        package_license = (
            package.license_declared.get_literals()[0].key
            if package.license_declared is not None and not isinstance(package.license_declared, SpdxNone)
            else ""
        )
        vendors = []
        related = []
        for refs in package.external_references:
            cpe_str = refs.locator
            cpe_parts = cpe_str.split(":")
            if cpe_parts[3] not in vendors:
                vendors.append(cpe_parts[3])
            if cpe_parts[4] not in related and cpe_parts[4] != doc.creation_info.name:
                related.append(cpe_parts[4])

        print(
            table_format.format(
                package_index, package.name, package.version, package_license, ", ".join(vendors), ", ".join(related)
            )
        )
        package_index += 1



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: 2018 Quentin Pradet https://gist.github.com/pquentin/5d8f5408cdad73e589d85ba509091741

import asyncio
import time

import aiohttp

START = time.monotonic()


class RateLimiter:
    """Rate limits an HTTP client that would make get() and post() calls.

    Calls are rate-limited by host.
    https://quentin.pradet.me/blog/how-do-you-rate-limit-calls-with-aiohttp.html
    This class is not thread-safe."""

    RATE = 1  # one request per second
    MAX_TOKENS = 10

    def __init__(self, client):
        self.client = client
        self.tokens = self.MAX_TOKENS
        self.updated_at = time.monotonic()

    async def get(self, *args, **kwargs):
        await self.wait_for_token()
        now = time.monotonic() - START
        print(f"{now:.0f}s: ask {args[0]}")
        return self.client.get(*args, **kwargs)

    async def wait_for_token(self):
        while self.tokens < 1:
            self.add_new_tokens()
            await asyncio.sleep(0.1)
        self.tokens -= 1

    def add_new_tokens(self):
        now = time.monotonic()
        time_since_update = now - self.updated_at
        new_tokens = time_since_update * self.RATE
        if self.tokens + new_tokens >= 1:
            self.tokens = min(self.tokens + new_tokens, self.MAX_TOKENS)
            self.updated_at = now


async def fetch_one(client, i):
    url = f"https://httpbin.org/get?i={i}"
    # Watch out for the extra 'await' here!
    async with await client.get(url) as resp:
        resp = await resp.json()
        now = time.monotonic() - START
        print(f"{now:.0f}s: got {resp['args']}")


async def main():
    async with aiohttp.ClientSession() as client:
        client = RateLimiter(client)
        tasks = [asyncio.ensure_future(fetch_one(client, i)) for i in range(20)]
        await asyncio.gather(*tasks)


if __name__ == "__main__":
    # Requires Python 3.7+
    asyncio.run(main())



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

import re

import semantic_version

from ics_sbom_libs.cve_match.package_matching.buildversion import BuildVersion
from ics_sbom_libs.cve_match.package_matching import VersionHandler, invalid_version_list


class OpenSSLVersionHandler(VersionHandler):
    def __init__(self):
        super(OpenSSLVersionHandler, self).__init__()
        self._package = "openssl"
        self._type = BuildVersion

    def convert(self, ver: str) -> str | semantic_version.Version:
        if ver in invalid_version_list:
            return ver

        parts = ver.split(".")
        if len(parts) not in [3, 4]:
            return ver

        offset = 0
        if len(parts) == 4:
            offset = 1

        major = parts[offset]
        minor = parts[offset + 1]
        patch = parts[offset + 2]

        # split patch to patch/build
        patch_parts = re.split(r"(\d+)", patch)
        patch = patch_parts[1]
        build = patch_parts[2]

        if offset != 1:
            pkg_version = self.version_type.coerce(f"{major}.{minor}.{patch}+{build}")
        else:
            pkg_version = self.version_type.coerce(f"{major}.{minor}.{patch}-{parts[0]}+{build}")

        return pkg_version



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

import semantic_version

from ics_sbom_libs.cve_match.package_matching.buildversion import BuildVersion
from ics_sbom_libs.cve_match.package_matching import VersionHandler, invalid_version_list


class LinuxKernelVersionHandler(VersionHandler):
    def __init__(self):
        super(LinuxKernelVersionHandler, self).__init__()
        self._package = "linux_kernel"
        self._type = BuildVersion

    def convert(self, ver: str) -> str | semantic_version.Version:
        if ver in invalid_version_list:
            return ver

        if " " in ver:
            ver = ver.strip()

        if ver.count("-") == 0:
            pkg_version = self.version_type.coerce(ver)
        else:
            parts = ver.split("-")
            if len(parts) != 2:

                return ver

            # for yocto platform type packages like linux-toradex the version of the kernel is before the '-'
            pkg_version = self.version_type.coerce(version_string=parts[0])

        return pkg_version



# SPDX-License-Identifier: LGPL-2.0-or-later
# SPDX-FileCopyrightText: 2024 Ics inc.
# SPDX-FileContributor: Michael Dingwall <mdingwall@ics.com>

from tqdm import tqdm
from ply import lex


class ProgressLexer:
    _progress: tqdm or None = None
    _lexer: lex or None = None

    def __init__(self, lexer):
        if lexer is not None:
            self._lexer = lexer

        self._last_pos = 0
        self._delta_pos = 0

    def build(self, **kwargs):
        self._lexer.build(**kwargs)

    def token(self):
        value = self._lexer.token()

        self._delta_pos = self._lexer.lexer.lexpos - self._last_pos
        self._last_pos = self._lexer.lexer.lexpos

        if self._progress is None:
            self._progress = tqdm(total=self._lexer.lexer.lexlen, desc="Reading SBoM", unit="b")

        self._progress.update(self._delta_pos)

        if value is None or value.type == "eof":
            self._progress.close()
        return value

    def input(self, data):
        self._lexer.input(data)
